/*
ping.c
*/

#define DEBUG	1

#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <net/gen/netdb.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <net/gen/oneCsum.h>
#include <fcntl.h>
#include <net/gen/in.h>
#include <net/gen/inet.h>
#include <net/gen/ip_hdr.h>
#include <net/gen/icmp_hdr.h>
#include <net/gen/ip_io.h>
#include <net/netlib.h>

#define WRITE_SIZE 		(30)
#define MTU_SZ			(576)

#define BUF_SZ 			(16*1024)
char buffer[BUF_SZ];

#if DEBUG
#define where() fprintf(stderr, "%s %d:", __FILE__, __LINE__);
#endif

/* */
static 
void 
	sig_hand(
		int sig
		){
}

static
void
	usage(
		char const *const argv0
		){

	fprintf(stderr, 
		/* [-t timeout] */
		"Usage: %s hostname [-I interface] [-l length]\n", 
		argv0
		);
	exit(1);
}

/* */
int
	main(
		int argc,
		char **argv
		){

	int fd, i, a;
	int result, result1;
	nwio_ipopt_t ipopt;
	ip_hdr_t *ip_hdr;
	int ihl;
	icmp_hdr_t *icmp_hdr;
	ipaddr_t dst_addr;
	struct hostent *hostent;

	char const *ip_device;
	unsigned length= WRITE_SIZE;

	/* */
	ip_device= getenv("IP_DEVICE");
	if(ip_device == NULL)ip_device= IP_DEVICE;

	/* parse cmdline */
	if(argc < 2)usage(argv[0]);
	for(a= 2; a < argc; ++a){

		if(argv[a][0] != '-')usage(argv[0]);
		if((a+1) >= argc)usage(argv[0]);

		switch(argv[a][1]){
		case 'I':
			++a;
			ip_device= argv[a];
			break;

		case 'l':
			++a;
			length= strtol (argv[a], (char **)0, 0);
			
			if(length < sizeof(icmp_hdr_t) + IP_MIN_HDR_SIZE){
				fprintf(stderr, "%s: length too small: %s\n",
					argv[0], argv[a]);
					exit(1);
			}
			if(length > MTU_SZ){ 
				fprintf(stderr, "%s: length too big: %s\n",
					argv[0], argv[a]);
					exit(1);
			}
			break;

		default:
			usage(argv[0]);
		}
	}

	/* */
	hostent= gethostbyname(argv[1]);
	if( hostent ){ 
		dst_addr= *(ipaddr_t *)(hostent->h_addr); 
	}else{
		dst_addr= inet_addr(argv[1]);
		if (dst_addr == -1)
		{
			fprintf(stderr, "%s: unknown host (%s)\n",
				argv[0], argv[1]);
			exit(1);
		}
	}
	fprintf(stderr,"dst host: %s\n", inet_ntoa(dst_addr) );

	/* */
	fprintf(stderr,"open -I: %s\n", ip_device);
	fd= open (ip_device, O_RDWR);
	if(fd<0){ perror("open"); exit(1); }

	ipopt.nwio_flags= NWIO_COPY | NWIO_PROTOSPEC;
	ipopt.nwio_proto= 1;

	result= ioctl (fd, NWIOSIPOPT, &ipopt);
	if (result<0)
		perror("ioctl (NWIOSIPOPT)"), exit(1);

	result= ioctl (fd, NWIOGIPOPT, &ipopt);
	if (result<0)
		perror("ioctl (NWIOGIPOPT)"), exit(1);

	fprintf(stderr, "ping length: %u\n", length);
	for (i= 0; i< 20; i++)
	{
		ip_hdr= (ip_hdr_t *)buffer;
		ip_hdr->ih_dst= dst_addr;

		icmp_hdr= (icmp_hdr_t *)(buffer+20);
		icmp_hdr->ih_type= 8;
		icmp_hdr->ih_code= 0;
		icmp_hdr->ih_chksum= 0;
		icmp_hdr->ih_chksum= ~oneC_sum(0, (u16_t *)icmp_hdr,
			WRITE_SIZE-20);
		result= write(fd, buffer, length);
		if (result<0)
		{
			perror("write");
			exit(1);
		}
		if (result != length)
		{
			where();
			fprintf(stderr, "result= %d\n", result);
			exit(1);
		}

		alarm(0);
		signal (SIGALRM, sig_hand);
		alarm(1);

		result= read(fd, buffer, sizeof(buffer));
		if (result>= 0 || errno != EINTR)
			break;
	}
	if (i >= 20)
	{
		printf("no answer from %s\n", argv[1]);
		exit(1);
	}
	if (result<0)
	{
		perror ("read");
		exit(1);
	}
	printf("%s is alive\n", argv[1]);
	exit(0);
}
