/* MINIX protocol A /dev/ttypXX filter */
#define _POSIX_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>

#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
/*#include <sys/select.h>*/

#define _MINIX
#include <termios.h>

/* system types */
typedef unsigned char 	uchar;
typedef unsigned short 	ushort;
typedef unsigned long 	ulong;
typedef unsigned int 	uint;
typedef char			bool;

#define _in
#define _out
#define _in_out

/* protocol A types */
#define PR_DLE 			(0x10U)
#define PR_SYN 			(0x16U)

#define PR_DLE_DLE 		(0x30U)
#define PR_DLE_SYN 		(0x31U)

	/* DLE-escaped aux char */
typedef int				t_prot_ch;
	/* application char */
typedef int				t_ch;

/* *************************************** */
/* cmdline prams */
static
bool
	/* is primary protocol A device (key -p was given) */
	is_pri= 0;

static
char
	/* auxXX_name[]= "/dev/tty00"; */
	*aux_name;

static
int
	/* argv position after [-r] */
	arg_i;

static
bool
	/* key [-r] was given */
	is_app_raw;

/* */
static
void
	/* 
	usage: -p|-s aux_name [-r] login_name [login_params ... ]
	"-p|-s" means primary|secondary "protocol A" device
	"-r" means ttypX in raw mode (stty raw </dev/ttypX)
	*/
	usage(
		void
		){

	fprintf(stderr,
		"usage: pra -p|-s aux_nm [-r] getty_nm [getty_params ... ]\n" 
		
		"  -p|-s\t\tpri|sec pra device\n"
		"  aux_nm\texternal aux port name (/dev/tty00)\n"
		"  -r\t\t(opt) raw ttypX (stty raw </dev/ttypX)\n"
		"  getty_nm\tttypX app name (nslogin|slip)\n"
		"  getty_params\t(opt) ttypX app params (/dev/psip1)\n"

		"examples:\n"
		"  pra -s /dev/tty01 nslogin\n"
		"  pra -s /dev/tty01 -r slip /dev/psip1\n"
		); 

	exit(1);
}

/* parse cmdline */
static
void
	parse_cmdline(
		int		argc,
		char 	**argv
		){

	if( argc < 4 ){ fprintf(stderr,"%s: no required parameters\n", "error"); usage(); }
	
	arg_i= 1;
	if(argv[arg_i][0] != '-'){ fprintf(stderr,"%s: -p|-s expected: %s\n", "error", argv[arg_i]); usage(); }
	switch(argv[arg_i][1]){ 
		case 's':
			is_pri= 0; ++arg_i;
			break;

		case 'p':
			is_pri= 1; ++arg_i;
			break;

		default:
			{ fprintf(stderr,"%s: -p|-s expected: %s\n", "error", argv[arg_i]); usage(); }
	}

	aux_name= argv[arg_i]; 
	++arg_i;
	
	is_app_raw= 0;
	if(argv[arg_i][0] == '-')
	switch(argv[arg_i][1]){ 
		case 'r':
			is_app_raw= 1; ++arg_i; 
			break;
		
		case '-':
			/* -- */ ++arg_i; 
			break;

		default:
			fprintf(stderr,"%s: unknown option: %s\n", "error", argv[arg_i]); 
			usage(); 
	}

	if(arg_i >= argc){ fprintf(stderr,"%s: no name to exec\n", "error"); usage(); }
}

/* *************************************** */
/* libmain */
static
pid_t
	/* PID of /dev/ttypX side (login session) */
	local_PID= 0;

static
void
	SIGINT_handler(
		int sig
		){
	
	/* signal(SIGINT,SIGINT_handler); */
	if(local_PID)kill(local_PID, SIGINT);
	
	exit(0);
}

/* */
static
void
	set_tty_raw(
		int aux_fd
		){

	struct termios
		t;
	
	if(tcgetattr(aux_fd, &t) == -1){ fprintf(stderr,"%s: tcsetattr AUX\n", "error"); perror("tcsetattr AUX"); exit(1); }
	/*
	memset(&t, 0, sizeof(t));
	t.c_ispeed= B9600;
	t.c_ospeed= B9600;
    */

	t.c_iflag= 0|IGNBRK;
	t.c_oflag= 0;
	t.c_cflag= 0|CLOCAL|CREAD|CS8;
	t.c_lflag= 0;
	
	cfsetispeed(&t,B115200);
	cfsetospeed(&t,B115200);

    /* wait new data if number of input chars is less then VMIN, 0 if no block input */
	t.c_cc[VMIN]=1;
    /* input wait timeout if number of chars is less then VMIN, 0 if no timeout */
	t.c_cc[VTIME]=0;

	if(tcsetattr(aux_fd, TCSANOW, &t) == -1){ fprintf(stderr,"%s: tcsetattr AUX\n", "error"); perror("tcsetattr AUX"); exit(1); }
}

/* *************************************** */
/* protocol A functions */

/* aux rx */
static
bool
	/* next aux_rx t_prot_ch must be second byte of DLE sequence */
	is_rx_DLE= 0;

/* 
	do DLE for AUX rx (protocol char -> data)
	return in (*p_rx_data) data to to write to pty
	return 0 if no data in (*p_rx_data) to write to pty (SYN or DLE)
*/
static
bool
	do_aux_rx_DLE(
		t_ch _out *p_rx_data,
		t_prot_ch _in rx_prot_ch
		){

	if(!p_rx_data){ fprintf(stderr,"%s: %s: %s\n", "error", "zptr", "do_aux_rx_DLE"); exit(1); }
    *p_rx_data= rx_prot_ch;

	/* ?expecting second DLE char from AUX */
	if(is_rx_DLE){
		
		switch(rx_prot_ch){
		case PR_DLE_DLE: *p_rx_data= PR_DLE; break;
		case PR_DLE_SYN: *p_rx_data= PR_SYN; break;
        /* 
        default: 
        invalid second DLE byte, 
        silently ignore error in the the implementation of protocol A
        and offer wrong char as is
        */
		}

		is_rx_DLE= 0;
		return 1;
	}
	
	/* check for first DLE char or SYN from AUX */
	switch(rx_prot_ch){ 
		case PR_DLE: is_rx_DLE= 1; return 0;
		case PR_SYN: return 0;
	}

	/* ordinary char */
	return 1;
}

/* aux_tx */
static
t_prot_ch
	/* second DLE protocol_char that is waiting for aux_tx */
	tx_second_DLE_prot_ch;

static
bool
	/* exist protocol_char symbol is waiting in (tx_second_char) */
	is_tx_DLE= 0;

/*
	{ tx_second_char, is_tx_DLE } aux_tx structure 
	can buffer only one t_ch data per time
	until the stored t_ch aux DLE transmission is not complete

	if you have other source of t_ch data instead of read_pty()
	to do_tx_DLE (t_ch -> t_prot_ch), you need 
		to provide other function instead of do_aux_tx_DLE()
		to fill { tx_second_char, is_tx_DLE } aux_tx structure 
	    
		and to correctly interleave calls of different versions 
		of do_aux_tx_DLE()

	for example:
	to post back incoming aux_rx not as echo protocol char, but as data, 
	you need to store the incoming aux_rx data into temporary data buffer 
	and then call own version do_aux_tx_DLE() for the temporary buffer

	to post back incoming aux_rx as echo protocol char, 
	if current tx_DLE in progress (is_tx_DLE != 0)
		you need complete current tx_DLE (post tx_next_char and discard echo)
		or discard current tx_DLE (is_tx_DLE= 0 and post echo)
*/

/*
	do_aux_tx_DLE(t_ch _in ch) version for: 
		do nonblock read from pty
		and do DLE for AUX tx (data -> protocol char)
	return (*p_tx_prot_ch) protocol_char was read from pty to write to aux_tx
	return 0 if no data in (*p_tx_prot_ch) to write to aux_tx (SYN)
*/
static
bool
	do_pty_in_DLE(
		t_prot_ch _out *p_tx_prot_ch,
		int _in pty_fd
		){
	
	int
		sz;

	char
		/* not 'int' because (ch) is partially init by read 
		and there is byte order in int */
		buf[1];

	if(!p_tx_prot_ch){ fprintf(stderr,"%s: %s: %s\n", "error", "zptr", "do_pty_in_DLE"); exit(1); }

	/* offer second byte to aux */
	if(is_tx_DLE){

		is_tx_DLE= 0;
		*p_tx_prot_ch= tx_second_DLE_prot_ch;
		return 1;
	}

	sz= read(pty_fd, buf, 1);
	/* no data to write to aux */
	if(sz != 1){ *p_tx_prot_ch= PR_SYN; return 0; }

	*p_tx_prot_ch= (*buf & 0x00FFU);

	/* test for start new DLE pair to aux */
	switch(*p_tx_prot_ch){
		default:
			/* ordinary char in tx_char to write to aux */
			return 1;

		case PR_DLE: tx_second_DLE_prot_ch= PR_DLE_DLE; break; 
		case PR_SYN: tx_second_DLE_prot_ch= PR_DLE_SYN; break; 
	}

	/* start new DLE pair in { tx_char, tx_second_char } to write to aux */
	*p_tx_prot_ch= PR_DLE; 
	is_tx_DLE= 1; 
	return 1;
}

/* 
	secondary:
	blocking read from AUX and do rx_DLE 
	store aux_rx data in rx_data
	return 0 if no data in (rx_data) to write to pty (SYN or DLE)
*/
static
bool
	sec_aux_rx_DLE(
		t_ch _out *p_rx_data,
		int _in aux_fd
		){

	t_prot_ch
		rx_prot_ch;

	int
		sz;

	char
		/* not 'int' because (*ptr) is partially init by read 
		and there is byte order in int */
		buf[1];

	if(!p_rx_data){ fprintf(stderr,"%s: %s: %s\n", "error", "zptr", "sec_aux_rx_DLE"); exit(1); }
	
	/* */
	sz= read(aux_fd, buf, 1);
    /* the blocking read should not return until data is ready,
    use suitable 'read' library call for the behaviour */
	if(sz == -1){ fprintf(stderr,"%s: read AUX\n", "EOF"); perror("read AUX"); exit(1); }
	
	rx_prot_ch= (*buf & 0x00FFU); 

    /* debug */
    /* *buf= prot_ch; write(2,buf,1); */

	return do_aux_rx_DLE(p_rx_data, rx_prot_ch);
}

/* primary */
static
bool
	/* primary main loop will not enter pause 
	because of tx_no_reply timeout */
	is_pri_tx_no_reply= 0;

/* 
	primary:
	non blocking read from AUX and do rx_DLE 
	store aux_rx data in rx_data
	extra parameter tx_prot_ch is used to display error messages better
	return 0 if no data in (rx_data) to write to pty (SYN or DLE)
*/
static
bool
	pri_aux_rx_DLE(
		t_ch _out *p_rx_data,
		int _in aux_fd,
		t_prot_ch _in tx_prot_ch
		){

	static
	ulong
		tx_to= 0;

	time_t
		st,
		et;

	t_prot_ch
		rx_prot_ch;

	int
		sz;

	char
		/* not 'int' because (*ptr) is partially init by read 
		and there is byte order in int */
		buf[1];

	if(!p_rx_data){ fprintf(stderr,"%s: %s: %s\n", "error", "zptr", "pri_aux_rx_DLE"); exit(1); }

	/* wait aux tx reply from secondary by NONBLOCK read aux */
	st= time(0);
	is_pri_tx_no_reply= 0;

	for(;;){
		errno= 0;
		sz= read(aux_fd, buf, 1); 
		if(sz == 1)break;

        /* try reduce spinlock */
		/*usleep(100);*/

		/* check no_reply timeout */
		et= time(0);
			/* pri_init_timer_no_reply */
		if((et - st) >= 2){

       		++tx_to;
       	    fprintf(stderr, "\raux tx 0x%02x: pri_tx_no_reply (0x%04x)", tx_prot_ch, tx_to); fflush(stderr);

       		is_pri_tx_no_reply= 1;
       		*p_rx_data= PR_SYN;
       		return 0;
		}

		/* break if do not wait for blocking read aux */
		/*fprintf(stderr, "drop wait aux input\n"); fflush(stderr);
		*buf= PR_SYN; break;*/

		if(errno == EAGAIN)continue;
		if(errno == EINTR)continue;
		{ fprintf(stderr,"%s: read AUX\n", "error"); perror("read AUX"); exit(1); }
	}

	rx_prot_ch= (*buf & 0x00FFU); 

    /* debug */
    /* *buf= prot_ch; write(2,buf,1); */

	return do_aux_rx_DLE(p_rx_data, rx_prot_ch);
}

/* *************************************** */
/* */
int
	main(
		int argc,
		char **argv
		){

/* number of ttypX allowed [ttyp0..ttyp<N>] */
#define ptypX_N 			(8)

	char
		/* digit N offset in string "/dev/ptypN" */
		#define ptypX_N_ofs (9)
		ptypX_name[16],
		/* digit N offset in string "/dev/ttypN" */
		#define ttypX_N_ofs (9)
		ttypX_name[16];

	uchar
		pty_n;

	int
		pty_fd,
		tmp;

	FILE
		*main_stderr;

	/* fdup_stderr */
	tmp= dup(2); if(tmp == -01){ fprintf(stderr,"dup stderr\n"); exit(1); }
	main_stderr= fdopen(tmp,"wt"); if(!main_stderr){ fprintf(stderr,"fdopen main_stderr\n"); exit(1); }
	
	if( setvbuf(stderr, 0, _IONBF, 0) ){ fprintf(stderr,"%s: setvbuf stderr\n", "error"); exit(1); }
	if( setvbuf(main_stderr, 0, _IONBF, 0) ){ fprintf(stderr,"%s: setvbuf main_stderr\n", "error"); exit(1); }

	parse_cmdline(argc,argv);

    /* try open /dev/ptypX */
	for(pty_n= 0; pty_n<ptypX_N; ++pty_n){

		sprintf( ptypX_name, "/dev/ptyp%01u", pty_n );
		sprintf( ttypX_name, "/dev/ttyp%01u", pty_n );

		/* open pty to NONBLOCK IO */
		pty_fd= open(ptypX_name, O_RDWR|O_NOCTTY|O_NONBLOCK );
		if(pty_fd != -1)break;

		perror(ptypX_name);
	}
	if(pty_n == ptypX_N){ fprintf(stderr,"enum /dev/ptypX failed\n"); exit(1); }

    /* print info */
	fprintf(stderr, "pra %s %s, pty: %s", 
		(is_pri? "-p": "-s"), aux_name, ptypX_name 
		); 
	fflush(stderr);

    /*ttyp:*/
	fprintf(stderr, ", %s%s", (is_app_raw? "-r ": ""), argv[arg_i] );
	if(argc-arg_i > 1)fprintf(stderr, " %s", argv[arg_i+1] );
	if(argc-arg_i > 2)fprintf(stderr, " ...[%u]", (argc-arg_i-2));
	fprintf(stderr, "\n");
	fflush(stderr);

    /* **** */
    /* local session */
	local_PID= fork();
	if(local_PID == -1){ fprintf(stderr,"%s: fork local\n", "error"); exit(1); }

    /* child: local session */
	if(!local_PID){
		int
			ttyp_fd;

		/* close parent (pty) side of ttyp */
		close(pty_fd);

		/* setsid must be called before open ttyp device (i.e. before dup2 0,1,2)
		perocess has no controlling terminal after setsid */
		if( setsid() == -1 ){ fprintf(main_stderr,"%s: setsid\n", "error"); exit(1); }
		
		/* exactly O_CTTY for open any new tty device
		will every time assign new controlling terminal to setsid process */
		ttyp_fd= open( ttypX_name, O_RDWR ); 
		if(ttyp_fd == -1){ fprintf(main_stderr,"%s: open %s\n", "error", ttypX_name); perror("open ttyp"); exit(1); }

		/* redirect 0,1,2 to ttyp */
		if( dup2(ttyp_fd, 0) == -1){ fprintf(main_stderr,"%s: dup2 0\n", "error"); exit(1); }
		if( dup2(ttyp_fd, 1) == -1){ fprintf(main_stderr,"%s: dup2 1\n", "error"); exit(1); }
		if( dup2(ttyp_fd, 2) == -1){ fprintf(main_stderr,"%s: dup2 2\n", "error"); exit(1); }

		/* ? ttypX in raw mode */
		if(is_app_raw){ set_tty_raw(ttyp_fd); }
		close(ttyp_fd);
		
		/* exec app_name [app_params] */
		execvp(argv[arg_i], &argv[arg_i]);

		/* execlp("getty", "getty", 0); */
		/* execlp("su","su","-e","ftp",0); */
		/* execlp("nslogin", "nslogin", 0); */
		/* execlp("slip", "slip", "/dev/psip1", 0); */
		{ fprintf(main_stderr,"%s: execXp\n", "error"); exit(1); }
	}

    /* else: server session */
	{
	int
		aux_fd;

	t_ch 
		rx_data;

	t_prot_ch 
		tx_prot_ch;

	/* ttyp user PID */
	/* fprintf(stderr,", PID: %u\n", local_PID); fflush(stderr); */
	signal(SIGINT, SIGINT_handler);

	/* *************************************** */
	/* pra -s (secondary) */
	if(!is_pri){ 
	
    /* open aux com device to blocking IO */
    aux_fd= open( aux_name, O_RDWR|O_NOCTTY );
	if(aux_fd == -1){ fprintf(stderr,"%s: open %s\n", "error", aux_name); perror("open aux"); exit(1); }

	/* aux to raw mode */
	set_tty_raw(aux_fd);
	/*set_tty_raw(pty_fd);*/

	/* pty NONBLOCK read version */
	for(;;){
		int
			sz;
		
		char
			/* beause (int) is partially init by read 
			and there is byte order in int */
			buf[1];

		/* 0 if aux read returned (SYN|DLE) in rx_data */
		if( sec_aux_rx_DLE(&rx_data, aux_fd) ){

			/* want rx_data blocking write pty */
			for(;;){
				*buf= rx_data;
				sz= write(pty_fd, buf, 1); 
				if(sz == 1)break;

				if(errno != EAGAIN){ fprintf(stderr,"%s: write pty\n", "error"); perror("write pty"); exit(1); }

    	        /* break if do not wait for blocking write pty */
        	    fprintf(stderr, "drop aux input 0x%02x\n", rx_data); fflush(stderr);
            	break;
			}
		}

		/* */
		/* want non_blocking read pty */
		do_pty_in_DLE(&tx_prot_ch, pty_fd);

		/* debug */
		/*if(tx_prot_ch != PR_SYN)*/
		{
			/* fputc(ch, stderr); / * fflush(stderr); */
			/* write(2,&tx,1); */
		}

		/* want tx_prot_ch blocking write aux */
		*buf= tx_prot_ch;
		sz= write(aux_fd, buf, 1);
		if(sz != 1){ fprintf(stderr,"%s: write AUX\n", "error"); perror("write AUX"); exit(1); }
	
	/* *************************************** */
	/* pra -p (primary) */
	}}else{
	ulong
		tx_pauses= 0;

    /* open aux com device to NONBLOCK IO */
    aux_fd= open( aux_name, O_RDWR|O_NOCTTY|O_NONBLOCK );
	if(aux_fd == -1){ fprintf(stderr,"%s: open %s\n", "error", aux_name); perror("open aux"); exit(1); }

	/* aux to raw mode */
	set_tty_raw(aux_fd);
	/*set_tty_raw(pty_fd);*/

	/*{ fprintf(stderr,"%s: not implemented: -p\n", "error"); exit(1); }*/
	for(;;){
		int
			sz;
		
		char
			/* beause (int) is partially init by read 
			and there is byte order in int */
			buf[1];
	
		/* want non_blocking read pty */
		do_pty_in_DLE(&tx_prot_ch, pty_fd);

		/* want tx_prot_ch blocking write aux */
		for(;;){
			*buf= tx_prot_ch;
			errno= 0;
			sz= write(aux_fd, buf, 1); 
			if(sz == 1)break;

			/* init_timer_tx_line */
			/* break if do not wait for blocking write pty */
            /*fprintf(stderr, "drop aux output 0x%02x\n", tx_prot_ch); fflush(stderr);
           	break;*/

			if(errno == EAGAIN)continue;
			if(errno == EINTR)continue;
			{ fprintf(stderr,"%s: write AUX\n", "error"); perror("write AUX"); exit(1); }
		}

		/* 0 if aux read returned (SYN|DLE) in rx_data */
		if( pri_aux_rx_DLE(&rx_data, aux_fd, tx_prot_ch) ){

			/* want rx_data blocking write pty */
			for(;;){
				*buf= rx_data;
				sz= write(pty_fd, buf, 1); 
				if(sz == 1)break;

				if(errno != EAGAIN){ fprintf(stderr,"%s: write pty\n", "error"); perror("write pty"); exit(1); }

    	        /* break if do not wait for blocking write pty */
        	    fprintf(stderr, "drop aux input 0x%02x\n", rx_data); fflush(stderr);
            	break;
			}
		}
		if( is_pri_tx_no_reply )continue;

		/* ? need enter pause step (detect SYN/SYN exchange) */
		if(tx_prot_ch == PR_SYN && rx_data == PR_SYN){

			++tx_pauses;
			fprintf(stderr, "\rpause (0x%08lx)", tx_pauses); fflush(stderr);
			
			/* pri_init_timer_keep_alive */
			usleep(1000);
		}
	}}

    /* end server session */
    }

	return 0;
}
